
from six import unichr
from nr.powerline.plugins import Plugin


class CharactersPlugin(Plugin):

  CHARACTERS = dict(
    # Icons
    USER = 0xf007,
    DIRECTORY = 0xf413,
    BRANCH = 0xe0a0,
    GIT_FOLDER = 0xe5fb,

    # Shapes
    RIGHT_TRIANGLE = 0xe0b0,
    LEFT_TRIANGLE = 0xe0b2,
    UPPER_LEFT_TRIANGLE = 0xe0bc,
    LOWER_LEFT_TRIANGLE = 0xe0b8,
    UPPER_RIGHT_TRIANGLE = 0xe0be,
    LOWER_RIGHT_TRIANGLE = 0xe0ba
  )

  def __getattr__(self, name):
    try:
      return unichr(self.CHARACTERS[name])
    except KeyError:
      raise AttributeError(name)




class Plugin(object):
  pass


def cached_property(func):
  cache_name = '_cached_' + func.__name__
  def cacher(self):
    try:
      return getattr(self, cache_name)
    except AttributeError:
      value = func(self)
      setattr(self, cache_name, value)
      return value
  def setter(self, value):
    setattr(self, cache_name, value)
  return property(cacher, setter)




BASH_PS1 = '''
function _nr_powerline_get() {
  echo "$PWD" | nc -U ~/.local/powerline/daemon.sock 2>/dev/null
  return $?
}
function _nr_powerline_alt() {
  if [ "$ALTPS1" != "" ]; then
    printf "$ALTPS1"
  else
    printf "$HOSTNAME $PWD $ "
  fi
}
function _nr_powerline_bootstrapper() {
  _nr_powerline_get
  if [ $? != 0 ]; then
    if which nr-powerline >/dev/null; then
      nr-powerline --start
      if ! ( nr-powerline --status --exit-code >/dev/null ); then
        echo "Could not start powerline daemon."
      else
        _nr_powerline_get
      fi
    else:
      _nr_powerline_alt
    fi
  fi
}
function _nr_powerline_direct() {
  nr-powerline 2>/dev/null
  if [ $? != 0 ]; then
    _nr_powerline_alt
  fi
}
if [[ `uname` =~ MSYS_NT.* ]]; then
  export PS1='`_nr_powerline_direct`'
else
  export PS1='`_nr_powerline_bootstrapper`'
fi
'''




from __future__ import print_function
from . import daemon
from .constants import BASH_PS1
from nr.utils.process import process_exists, process_terminate, replace_stdio, spawn_daemon
import argparse
import os
import signal
import sys


def default_powerline():
  from nr.powerline import PowerLine
  p = PowerLine()
  git = p.get_plugin('git')

  p.set_pen('white', 'blue')
  p.add_part(' {c.GIT_FOLDER} ' if git.project else ' {c.DIRECTORY} ')
  p.add_part('{session.cwd} !{c.RIGHT_TRIANGLE}')
  if git.project:
    p.set_pen(None, 'yellow')
    p.add_part(' {c.BRANCH} {git.branch} !{c.RIGHT_TRIANGLE}')
  p.add_part(' ')
  p.clear_pen()
  return p


def powerline_supplier(filename):
  if filename:
    with open(filename) as fp:
      code = compile(fp.read(), 'exec', filename)
    def func():
      scope = {}
      exec(code, scope)
      return scope['powerline']
  else:
    func = default_powerline
  return func



def main(argv=None, prog=None):
  parser = argparse.ArgumentParser(prog=prog)
  parser.add_argument('-f', '--file')
  parser.add_argument('--run-dir', default=None)
  parser.add_argument('--start', action='store_true')
  parser.add_argument('--stop', action='store_true')
  parser.add_argument('--status', action='store_true')
  parser.add_argument('--exit-code', action='store_true')
  parser.add_argument('--src', choices=('bash',))
  args = parser.parse_args(argv)

  if args.src == 'bash':
    print(BASH_PS1)
    sys.exit(0)
  elif args.src:
    parser.error('unexpected argument for --src: {!r}'.format(args.src))

  if not args.start and not args.stop and not args.status:
    powerline_supplier(args.file)().print_()
    return

  run_dir = args.run_dir or os.path.expanduser('~/.local/powerline')
  log_file = os.path.join(run_dir, 'daemon.log')
  pid_file = os.path.join(run_dir, 'daemon.pid')
  socket_file = os.path.join(run_dir, 'daemon.sock')

  if os.path.isfile(pid_file):
    with open(pid_file) as fp:
      daemon_pid = int(fp.read().strip())
  else:
    daemon_pid = None

  if args.stop and daemon_pid:
    print('Stopping', daemon_pid)
    process_terminate(daemon_pid)
  if args.start:
    def run(powerline, stdout):
      with open(pid_file, 'w') as fp:
        fp.write(str(os.getpid()))
      print('Started', os.getpid())
      signal.signal(signal.SIGTERM, lambda: os.remove(pid_file))
      replace_stdio(None, stdout, stdout)
      conf = daemon.SocketConf.UnixFile(socket_file)
      daemon.PowerlineDaemon(conf, powerline).run_forever()

    powerline = powerline_supplier(args.file)
    os.makedirs(run_dir, exist_ok=True)
    stdout = open(log_file, 'a')
    spawn_daemon(lambda: run(powerline, stdout))
  if args.status:
    if not daemon_pid or not process_exists(daemon_pid):
      if args.exit_code:
        sys.exit(7)
      print('stopped')
    else:
      if args.exit_code:
        sys.exit(0)
      print('running')



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from __future__ import print_function

import copy
import os
import pkg_resources
import nr.parsing.core
import nr.sumtype
import re
import six
import sys
import termcolor

__author__ = 'Niklas Rosenstein <rosensteinniklas@gmail.com>'
__version__ = '0.0.5'

PLUGINS_ENTRYPOINT = 'nr.powerline.plugins'
COLORMODE_ASCII = 'ascii'
COLORMODE_TRUECOLOR = 'truecolor'


@six.python_2_unicode_compatible
class PowerLine(object):

  def __init__(self):
    self._parts = []
    self._plugins = {}
    self._colors = {}
    self._color_mode = COLORMODE_ASCII
    self._pen = Pen()
    self._drawops = []

  def __str__(self):
    result = [termcolor.RESET]
    pen = Pen()
    new_pen = pen
    for i, op in enumerate(self._drawops):
      if op.is_set_pen():
        new_pen = op.pen
      elif op.is_propagate_bg():
        # Look ahead for the next foreground color.
        new_pen = pen.copy()
        new_pen.fg = new_pen.bg
        new_pen.attrs = []
        for anop in self._drawops[i:]:
          if anop.is_set_pen():
            new_pen.bg = anop.pen.bg
            break
        else:
          new_pen.bg = None
      elif op.is_blit_text():
        result.append(render_format_string(op.text, _LazyPluginDict(self)))
      else:
        assert False, op
      if pen != new_pen:
        result.append(self.get_color_transition_codes(pen, new_pen))
        pen = new_pen
    result.append(termcolor.RESET)
    return u''.join(result)

  @property
  def color_mode(self):
    return self._color_mode

  @color_mode.setter
  def color_mode(self, value):
    assert value in (COLORMODE_ASCII, COLORMODE_TRUECOLOR), repr(value)
    self._color_mode = value

  @property
  def colors(self):
    return _ColorDict(self._colors)

  def add_color(self, name, value):
    self.colors[name] = value

  def get_plugin(self, name):
    if name not in self._plugins:
      for ep in pkg_resources.iter_entry_points(PLUGINS_ENTRYPOINT, name):
        self._plugins[name] = ep.load()()
        break
      else:
        return None
    return self._plugins[name]

  def set_pen(self, fg=NotImplemented, bg=NotImplemented, *attrs):
    self._pen.update(fg, bg, attrs)
    self._drawops.append(DrawOp.SetPen(self._pen.copy()))

  def clear_pen(self):
    self._pen.clear()

  def add_part(self, part_format):
    scanner = nr.parsing.core.Scanner(part_format)
    start = scanner.index
    while scanner:
      if scanner.char == '!':
        self._drawops.append(DrawOp.BlitText(part_format[start:scanner.index]))
        start = scanner.index + 1
        self._drawops.append(DrawOp.PropagateBg())
      scanner.next()
    self._drawops.append(DrawOp.BlitText(part_format[start:scanner.index]))

  def get_color_code(self, name, foreground=True):
    if self._color_mode == COLORMODE_TRUECOLOR and name in self._colors:
      # TODO @NiklasRosenstein
      return ''
    mapping = termcolor.COLORS if foreground else termcolor.HIGHLIGHTS
    prefix = '' if foreground else 'on_'
    if prefix + name not in mapping:
      return ''
    return '\033[%dm' % mapping.get(prefix + name)

  def get_color_transition_codes(self, curr, pen):
    # TODO @NiklasRosenstein We should be able to compute only the terminal
    #   instructions to change to the new colors instead of re-writing all
    #   color instructions.
    result = termcolor.RESET
    if pen.fg:
      result += self.get_color_code(pen.fg, True)
    if pen.bg:
      result += self.get_color_code(pen.bg, False)
    if pen.attrs:
      for attr in pen.attrs:
        result += '\033[%dm' % termcolor.ATTRIBUTES.get(attr, '')
    return result

  def print_(self):
    sys.stdout.buffer.write(str(self).encode('utf8'))
    #if os.name == 'nt':
    #else:
    #  print(str(self), end='')


class Pen(object):

  def __init__(self, fg=None, bg=None, attrs=None):
    self.fg = fg
    self.bg = bg
    self.attrs = attrs or []

  def __repr__(self):
    return 'Pen(fg={!r}, bg={!r}, attrs={!r})'.format(self.fg, self.bg, self.attrs)

  def __eq__(self, other):
    if type(other) is Pen:
      return (self.fg, self.bg, self.attrs) == (other.fg, other.bg, other.attrs)
    return False

  def __ne__(self, other):
    return not (self == other)

  def update(self, fg=NotImplemented, bg=NotImplemented, attrs=NotImplemented):
    if fg is not NotImplemented:
      self.fg = fg
    if bg is not NotImplemented:
      self.bg = bg
    if attrs is not NotImplemented:
      if attrs is None:
        attrs = []
      elif not isinstance(attrs, (list, tuple)):
        attrs = [attrs]
      self.attrs = attrs

  def clear(self):
    self.fg = None
    self.bg = None
    self.attrs = []

  def copy(self):
    return copy.deepcopy(self)


class TrueColor(object):

  def __init__(self, value):
    if isinstance(value, str):
      if value.startswith('#'):
        v = value[1:]
        if len(v) == 3:
          r, g, b = v
          r *= 2
          g *= 2
          b *= 2
        elif len(v) == 6:
          r, g, b = v[0:2], v[2:4], v[4:6]
        else:
          raise ValueError('invalid RGB color: {!r}'.format(value))
        r, g, b = int(r, 16), int(g, 16), int(b, 16)
      else:
        raise ValueError('invalid RGB color: {!r}'.format(value))
    elif isinstance(value, Color):
      r, g, b = value.r, value.g, value.b
    else:
      raise TypeError('unexpected type {}'.format(type(value).__name__))
    self.r = r
    self.g = g
    self.b = b

  def __repr__(self):
    return 'TrueColor(r={}, g={}, b={})'.format(self.r, self.g, self.b)


class _ColorDict(object):

  def __init__(self, colors):
    self._colors = colors

  def __getitem__(self, name):
    return self._colors[name]

  def __setitem__(self, name, value):
    self._colors[name] = TrueColor(value)

  def __delitem__(self, name):
    del self._colors[name]


class _LazyPluginDict(object):

  def __init__(self, powerline):
    self._powerline = powerline

  def __getitem__(self, key):
    plugin = self._powerline.get_plugin(key)
    if plugin is None:
      raise KeyError(key)
    return plugin


@nr.sumtype.add_constructor_tests
class DrawOp(nr.sumtype.Sumtype):
  PropagateBg = nr.sumtype.Constructor()  # Indicated by the "!" character in a format string
  SetPen = nr.sumtype.Constructor('pen')
  BlitText = nr.sumtype.Constructor('text')


def render_format_string(fmt_string, values):
  """
  Similar to :meth:`str.format`. Can read from arbitrary dictionary-like
  objects. Replaces non-existent values with the same syntax string.
  """

  expr = '\{(.*?)\}'

  def subst_func(m):
    parts = m.group(1).split('.')
    try:
      value = values[parts[0]]
    except KeyError:
      return '{' + m.group(1) + '}'
    for part in parts[1:]:
      try:
        value = getattr(value, part)
      except AttributeError:
        return '{' + m.group(1) + '}'
    return six.text_type(value)

  return re.sub(expr, subst_func, fmt_string)




import os
import re
import subprocess

from nr.powerline.plugins import Plugin, cached_property


def get_output(command, default=''):
  try:
    data = subprocess.check_output(
      command,
      shell=True,
      stderr=open(os.devnull, 'w'))
    return data.decode().strip()
  except subprocess.CalledProcessError:
    return default


class GitPlugin(Plugin):

  @cached_property
  def user(self):
    return get_output('git config user.email')

  @cached_property
  def project(self):
    return os.path.basename(get_output('git rev-parse --show-toplevel'))

  @cached_property
  def branch(self):
    branches = [x.strip() for x in re.split('\s+', get_output('git branch'))]
    try:
      index = branches.index('*')
    except ValueError:
      return 'master'
    else:
      return branches[index+1]




import os
import platform

from nr.powerline.plugins import Plugin, cached_property


class SessionPlugin(Plugin):

  _cwd = None

  @cached_property
  def hostname(self):
    return platform.node()

  @property
  def cwd(self):
    user_home = os.path.expanduser('~')
    cwd = self._cwd or os.getcwd()
    try:
      relpath = os.path.relpath(cwd, user_home)
      if relpath.startswith(os.pardir):
        raise ValueError
      if relpath == os.curdir:
        return '~'
      return '~/' + relpath
    except ValueError:
      return cwd
    assert False



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from . import PowerLine
import socket as _socket
import nr.sumtype
import os
import traceback
from typing import Callable


class SocketConf(nr.sumtype.Sumtype):
  Ipv4 = nr.sumtype.Constructor('host,port')
  UnixFile = nr.sumtype.Constructor('filename')

  @Ipv4.member
  def type(self) -> int:
    return _socket.AF_INET

  @UnixFile.member
  def type(self) -> int:
    return _socket.AF_UNIX

  @Ipv4.member
  def bind(self, socket: _socket.socket):
    socket.bind((self.host, self.port))

  @Ipv4.member
  def connect(self, socket: _socket.socket):
    socket.connect((self.host, self.port))

  @UnixFile.member
  def bind(self, socket: _socket.socket):
    socket.bind(self.filename)

  @UnixFile.member
  def connect(self, socket: _socket.socket):
    socket.connect(self.filename)


class PowerlineDaemon:

  def __init__(self, conf: SocketConf, powerline: Callable[[], PowerLine]) -> None:
    self._conf = conf
    self._powerline = powerline

  def run_forever(self):
    socket = _socket.socket(self._conf.type())
    self._conf.bind(socket)
    socket.listen(5)
    socket.settimeout(0.1)
    try:
      while True:
        try:
          conn, address = socket.accept()
        except _socket.timeout:
          continue
        path = conn.makefile().readline().strip()
        try:
          os.chdir(path)
          result = str(self._powerline())
        except Exception as exc:
          result = traceback.format_exc()
        conn.makefile('w').write(result)
        conn.close()
    finally:
      socket.close()
      if isinstance(self._conf, SocketConf.UnixFile):
        os.remove(self._conf.filename)


class PowerlineClient:

  def __init__(self, conf: SocketConf):
    self._conf = conf

  def request(self, path: str) -> str:
    socket = _socket.socket(self._conf.type())
    self._conf.connect(socket)
    path = os.path.abspath(path)
    try:
      socket.makefile('w').write(path + '\n')
      return socket.makefile().read()
    finally:
      socket.close()



# automatically created by shore 0.0.22

import io
import re
import setuptools
import sys

with io.open('src/nr/powerline/__init__.py', encoding='utf8') as fp:
  version = re.search(r"__version__\s*=\s*'(.*)'", fp.read()).group(1)

with io.open('README.md', encoding='utf8') as fp:
  long_description = fp.read()

requirements = ['nr.parsing.core >=0.0.1,<0.1.0', 'nr.sumtype >=0.0.3,<0.1.0', 'nr.utils.process >=0.0.3,<0.1.0', 'termcolor >=1.1.0,<2.0.0']

setuptools.setup(
  name = 'nr.powerline',
  version = version,
  author = 'Niklas Rosenstein',
  author_email = 'rosensteinniklas@gmail.com',
  description = 'Package description here.',
  long_description = long_description,
  long_description_content_type = 'text/markdown',
  url = None,
  license = 'MIT',
  packages = setuptools.find_packages('src', ['test', 'test.*', 'docs', 'docs.*']),
  package_dir = {'': 'src'},
  include_package_data = True,
  install_requires = requirements,
  extras_require = {},
  tests_require = [],
  python_requires = None, # TODO: '>=3.4,<4.0.0',
  data_files = [],
  entry_points = {
    'console_scripts': [
      'nr-powerline = nr.powerline.main:main',
    ],
    'nr.powerline.plugins': [
      'c = nr.powerline.plugins.characters:CharactersPlugin',
      'git = nr.powerline.plugins.git:GitPlugin',
      'session = nr.powerline.plugins.session:SessionPlugin',
      'time = nr.powerline.plugins.time:TimePlugin',
    ]
  },
  cmdclass = {},
  keywords = [],
  classifiers = [],
)




import datetime

from nr.powerline.plugins import Plugin, cached_property


class TimePlugin(Plugin):

  @cached_property
  def now(self):
    return datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
